"""Stela Utils module."""
from enum import Enum, unique
from importlib import reload
from pathlib import Path
from typing import Optional


@unique
class StelaFileType(Enum):
    """Available formats for configuration files."""

    INI = [".ini"]
    JSON = [".json"]
    YAML = [".yaml", ".yml"]
    TOML = [".toml"]


def stela_reload() -> "StelaCut":  # type: ignore
    """Reload Stela configuration.

    This helper will import again stela module
    to update settings.

    """
    import stela

    reload(stela)
    return stela.settings


def find_pyproject_folder() -> Optional[Path]:
    """Find pyproject.toml file.

    Look for:
        1. Current working directory up to root
        2. Current file directory up to root


    :return: Optional Path
    """

    def look_for_file(current_path: Path) -> Optional[Path]:
        if current_path.joinpath("pyproject.toml").exists():
            return current_path
        if str(current_path) in ["/", "\\"]:
            return None
        return look_for_file(current_path.parent)

    found_path = look_for_file(Path().cwd())
    if found_path:
        return found_path

    found_path = look_for_file(Path(__file__))
    return found_path



"""Stela Options module."""
import os
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import toml

from stela.exceptions import StelaEnvironmentNotFoundError, StelaFileTypeError
from stela.utils import StelaFileType, find_pyproject_folder


@dataclass
class StelaOptions:
    """Stela Options data class."""

    current_environment: Optional[str] = None
    default_environment: Optional[str] = None
    environment_variable_name: str = "ENVIRONMENT"
    config_file_prefix: str = ""
    config_file_suffix: str = ""
    environment_prefix: str = ""
    environment_suffix: str = ""
    config_file_extension: StelaFileType = StelaFileType.INI
    evaluate_data: bool = False
    config_file_path: str = "."
    filenames: List[str] = field(default_factory=list)
    show_logs: bool = True
    do_not_read_environment: bool = False

    def get_extensions(self) -> List[str]:
        """Return file extensions for project configuration files."""
        return self.config_file_extension.value  # type: ignore

    @classmethod
    def get_from_env_or_settings(
        cls, key: str, file_settings: Dict[str, Any], default: Any
    ) -> Any:
        """Get from environment or settings the Stela Option argument.

        :param key: Current stela argument.
        :param file_settings: Stela Data from pyproject.toml.
        :param default: default value for argument.
        :return: Any
        """
        environment_name = f"STELA_{key.upper().replace('.', '_')}"
        return os.getenv(environment_name, file_settings.get(key, default))

    @classmethod
    def get_config(cls) -> "StelaOptions":
        """Get config from pyproject.toml."""

        file_settings = {}
        path = find_pyproject_folder()
        if path:
            filepath = path.joinpath(f"pyproject.toml")
            toml_settings = toml.load(filepath)
            file_settings = toml_settings.get("tool", {}).get("stela", {})
        settings = {
            "environment_variable_name": cls.get_from_env_or_settings(
                "environment_variable_name",
                file_settings,
                cls.environment_variable_name,
            ),
            "default_environment": cls.get_from_env_or_settings(
                "default_environment", file_settings, cls.default_environment
            ),
            "config_file_prefix": cls.get_from_env_or_settings(
                "config_file_prefix", file_settings, cls.config_file_prefix
            ),
            "config_file_suffix": cls.get_from_env_or_settings(
                "config_file_suffix", file_settings, cls.config_file_suffix
            ),
            "environment_prefix": cls.get_from_env_or_settings(
                "environment_prefix", file_settings, cls.environment_prefix
            ),
            "environment_suffix": cls.get_from_env_or_settings(
                "environment_suffix", file_settings, cls.environment_suffix
            ),
            "evaluate_data": bool(
                cls.get_from_env_or_settings(
                    "evaluate_data", file_settings, cls.evaluate_data
                )
            ),
            "config_file_path": cls.get_from_env_or_settings(
                "config_file_path", file_settings, cls.config_file_path
            ),
            "show_logs": cls.get_from_env_or_settings(
                "show_logs", file_settings, cls.show_logs
            ),
            "do_not_read_environment": cls.get_from_env_or_settings(
                "do_not_read_environment", file_settings, cls.do_not_read_environment
            ),
        }
        try:
            config_file_extension = cls.get_from_env_or_settings(
                "config_file_extension",
                file_settings,
                cls.config_file_extension.value[0].replace(".", "").upper(),
            )
            settings["config_file_extension"] = StelaFileType[config_file_extension]
        except KeyError:
            raise StelaFileTypeError(
                f"Invalid file type: {file_settings.get('config_file_extension')}"
            )
        settings["current_environment"] = os.getenv(
            settings["environment_variable_name"]
        )
        if not settings["current_environment"]:
            settings["current_environment"] = settings["default_environment"]
        if not settings["current_environment"]:
            raise StelaEnvironmentNotFoundError(f"Environment not found")
        settings["filenames"] = [
            f"{settings['config_file_prefix']}{settings['current_environment']}{extension}"
            for extension in settings["config_file_extension"].value
        ]
        return cls(**settings)



"""Import Stela Data.

If you need to reload data inside you application,
use the stela_reload function.

"""
from loguru import logger

from stela.stela import Stela
from stela.stela_cut import StelaCut
from stela.stela_options import StelaOptions
from stela.utils import stela_reload

settings: StelaCut = Stela(
    stela_options=StelaOptions.get_config()
).get_project_settings()

if not settings.stela_options.show_logs:
    logger.disable("stela")



"""Stela Class.

This class run the Stela lifecycle:
   Pre-Load (optional) -> Load or Default Loader -> Post-Load (optional)
"""
import configparser
import json
from pathlib import Path
from typing import Any, Dict

import toml
import yaml

from stela.exceptions import StelaEnvironmentNotFoundError
from stela.stela_cut import StelaCut
from stela.stela_options import StelaOptions
from stela.utils import find_pyproject_folder


class Stela:
    """Stela Class."""

    def __init__(self, stela_options: StelaOptions) -> None:
        """Initialize class.

        :param stela_options: StelaOptions instance
        """
        self.options = stela_options

    def default_loader(
        self, data: Dict[Any, Any], options: StelaOptions
    ) -> Dict[Any, Any]:
        """Stela Default Loader.

        :param data: Current data parsed from pre-load
        :param options: StelaOptions instance
        :return: Dict
        """
        from loguru import logger

        path = find_pyproject_folder() or Path().cwd()
        for filename in options.filenames:
            filepath = path.joinpath(self.options.config_file_path, filename)
            logger.debug(f"Looking for file {filepath}...")
            if filepath.exists():
                settings_data = self.load_from_file(filepath)
                data.update(settings_data)
                return data
        return data

    def get_project_settings(self) -> "StelaCut":
        """Get project settings running Stela Lifecycle.

        :return: Dict
        """
        settings_data = {}

        # Run pre_load
        if getattr(self.options, "pre_load", None) is not None:
            pre_load_data = self.options.pre_load(options=self.options)  # type: ignore
            settings_data.update(pre_load_data)

        # Run load or default_load
        if getattr(self.options, "load", None) is not None:
            load_data = self.options.load(data=settings_data, options=self.options)  # type: ignore
        else:
            load_data = self.default_loader(data=settings_data, options=self.options)
        settings_data.update(load_data)

        # Run post_load
        if getattr(self.options, "post_load", None) is not None:
            pre_load_data = self.options.post_load(  # type: ignore
                data=settings_data, options=self.options
            )
            settings_data.update(pre_load_data)

        proxy = StelaCut(settings_data)
        proxy.stela_options = self.options
        return proxy

    @property
    def environment(self) -> str:
        """Return Current Environment."""
        if not self.options.current_environment:
            raise StelaEnvironmentNotFoundError("Environment not found.")
        return self.options.current_environment

    def load_from_file(self, filepath: Path) -> Dict[Any, Any]:
        """Resolve correct function for file extension."""
        function_name = (
            f"load_{self.options.config_file_extension.value[0].replace('.', '')}"
        )
        return getattr(self, function_name)(filepath)  # type: ignore

    def load_ini(self, filepath: Path) -> Dict[Any, Any]:
        """Load INI files.

        :param filepath: Path instance
        :return: Dict
        """
        config = configparser.ConfigParser()
        config.read(filepath)
        ini_settings: Dict[Any, Any] = {}
        for main_key in config.keys():
            ini_settings[main_key] = {}
            for key in config[main_key].keys():
                ini_settings[main_key][key] = config[main_key][key]
        return ini_settings

    def load_yaml(self, filepath: Path) -> Dict[Any, Any]:
        """Load YAML files.

        :param filepath: Path instance
        :return: Dict
        """
        with open(filepath, "r") as yaml_file:
            config = yaml.safe_load(yaml_file)
            return config  # type: ignore

    def load_toml(self, filepath: Path) -> Dict[Any, Any]:
        """Load TOML files.

        :param filepath: Path instance
        :return: Dict
        """
        config = toml.load(filepath)
        return config  # type: ignore

    def load_json(self, filepath: Path) -> Dict[Any, Any]:
        """Load JSON files.

        :param filepath: Path instance
        :return: Dict
        """
        with open(filepath, "r") as json_file:
            config = json.load(json_file)
            return config  # type: ignore



"""Stela Decorators."""
from functools import wraps


def pre_load(f):
    """Pre-Loader decorator.

    Use this decorator to handle data
    before invoking
    the custom load or default_loader.

    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.stela_options import StelaOptions

    if not getattr(StelaOptions, "pre_load", None):
        setattr(StelaOptions, "pre_load", f)

    return wrapper


def load(f):
    """Run Custom Loader decorator.

    This loader will be invoked instead
    the default_loader.

    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.stela_options import StelaOptions

    if not getattr(StelaOptions, "load", None):
        setattr(StelaOptions, "load", f)

    return wrapper


def post_load(f):
    """Post-Loader decorator.

    Use this decorator to handle data
    after invoking
    the custom load or default_loader.
    
    :param f: decorated function
    :return: wrapper function
    """

    @wraps(f)
    def wrapper(self, *args, **kwargs):
        return f(self, *args, **kwargs)

    from stela.stela_options import StelaOptions

    if not getattr(StelaOptions, "post_load", None):
        setattr(StelaOptions, "post_load", f)

    return wrapper



"""Stela Cut subclass.

Based on Scalpl project: https://github.com/ducdetronquito/scalpl

"""
import os
from ast import literal_eval
from typing import Any, Dict, Optional, Tuple

from scalpl import Cut
from scalpl.scalpl import index_error, key_error, split_path, traverse, type_error


class FakeNone:
    """Fake None for Stela Cut.

    Resolves StelaCut.cut() and StelaCut.pop()
    issues when default=None.
    See: https://github.com/ducdetronquito/scalpl/issues/16
    """

    pass


class StelaCut(Cut):  # type: ignore
    """StelaCut subclass."""

    @property
    def stela_options(self) -> "StelaOptions":  # type: ignore
        """Return Stela Options."""
        return self._options

    @stela_options.setter
    def stela_options(self, options: "StelaOptions"):  # type: ignore
        """Save Stela Options in instance."""
        self._options = options

    @property
    def to_dict(self) -> Dict[Any, Any]:
        """Return Cut as Python Dictionary.

        When thinking on Scalpl context, its fair to
        convert data from Cut to dict.

        But in Stela, the main context is to use the full
        path key to investigate possible environment variables
        before returning the combined keys' value.

        For example, suppose we have:
        APP = "my project" and
        APP_SECRET = "bar"

        Our dictionary:
        {
            "app": {
                "secret": "foo"
                }
            }
        }

        Retrieving settings["app.secret"] can reach both values (dict and env) correctly.
        Retrieving settings["app"] correct gives us the env value.

        But if you're looking only at the Dict, determine both
        environment variables can be tricky.

        We suggest use this property for dictionary-only contexts
        or for tests.
        """
        return dict(self)

    def get(self, path: str, default: Optional[Any] = FakeNone) -> Any:
        """Override Get from Dict code."""
        try:
            return self[path]
        except (KeyError, IndexError) as error:
            if default is not FakeNone:
                return default
            raise error

    def pop(self, path: str, default: Any = FakeNone, *args) -> Any:  # type: ignore
        """Override Pop from Dict code."""
        *keys, last_key = split_path(path, self.sep)

        try:
            item = traverse(data=self.data, keys=keys, original_path=path)
        except (KeyError, IndexError) as error:
            if default is not FakeNone:
                return default
            raise error

        try:
            return item.pop(last_key)
        except (AttributeError, KeyError) as error:
            if default is not FakeNone:
                return default
            raise key_error(last_key, path, error)
        except IndexError as error:
            if default is not FakeNone:
                return default
            raise index_error(last_key, path, error)

    def __getitem__(
        self, path: str, *args: Tuple[Any], **kwargs: Dict[Any, Any]
    ) -> Any:
        """Override getitem from Dict.

        Before invoking  original code, we will
        use the full path received to check for
        a environment variable value, using StelaOptions
        options.

        :param path: full dict path
        :param args: python arguments
        :param kwargs: python keyword arguments
        :return: Any
        """
        if not "[" in path and not self.stela_options.do_not_read_environment:
            environment_variable = (
                f"{self.stela_options.environment_prefix}"
                f"{path.replace('.', '_')}"
                f"{self.stela_options.environment_suffix}".upper()
            )
            value = os.getenv(environment_variable)
            if value:
                from loguru import logger

                logger.debug(f"Using value from variable {environment_variable}")
                if self.stela_options.evaluate_data:
                    try:
                        value = literal_eval(value)
                    except:
                        pass
                return value

        *keys, last_key = split_path(path, self.sep)
        item = traverse(data=self.data, keys=keys, original_path=path)

        try:
            return item[last_key]
        except KeyError as error:
            raise key_error(last_key, path, error)
        except IndexError as error:
            raise index_error(last_key, path, error)
        except TypeError:
            raise type_error(last_key, path, item)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stela']

package_data = \
{'': ['*']}

install_requires = \
['loguru', 'pyyaml', 'scalpl', 'toml']

extras_require = \
{':python_version < "3.7"': ['dataclasses']}

setup_kwargs = {
    'name': 'stela',
    'version': '1.0.12',
    'description': 'Find and read your project configuration files easily',
    'long_description': '# Welcome to Stela\n\n[![PyPI](https://img.shields.io/pypi/v/stela)](https://pypi.org/project/stela/)\n[![Build](https://github.com/chrismaille/stela/workflows/tests/badge.svg)](https://github.com/chrismaille/stela/actions)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/stela)](https://www.python.org)\n[![Conventional Commits](https://img.shields.io/badge/Conventional%20Commits-1.0.0-yellow.svg)](https://conventionalcommits.org)\n<a href="https://github.com/psf/black"><img alt="Code style: black"\nsrc="https://img.shields.io/badge/code%20style-black-000000.svg"></a>\n\n[Stela](https://en.wikipedia.org/wiki/Stele) were the "configuration\nfiles" of ancient times. This library aims to simplify your project\nconfigurations, from *json*, *yaml*, *ini* or *toml* files in a single\nPython dictionary, which values can be easily be override by environment\nvariables.\n\n### Install\n\n```shell\n$ pip install stela\n```\n\n### Example\n\nYou can use `.yaml`, `.ini`, `.json` and `.toml` files. Suppose a file\ncalled `development.ini` which contains:\n\n```ini\n[foo]\nbar = value\n```\n\nStela will convert data into a python dictionary:\n\n```python\n{\n    "foo": {\n        "bar": "value"\n    }\n}\n```\n\nAnd you can use the settings like this:\n\n```python\nfrom stela import settings\n\nmy_conf = settings["foo.bar"]\n# my_conf = "value"\n```\n\nThis is possible because Stela uses under the hood the\n[Scalpl](https://github.com/ducdetronquito/scalpl) library.\n\n### How Stela find the configuration files?\n\nBy default, Stela will use the value from `ENVIRONMENT` environment\nvariable to find correspondent INI file. For example, if you set\n`ENVIRONMENT=development` Stela will look for `development.ini` file in\nproject root.\n\nYou can change this behaviour inside `pyproject.toml` file:\n\n```toml\n[tools.stela]\nenvironment_variable_name = "ENVIRONMENT"\nconfig_file_extension = "INI" # YAML, TOML, JSON\nconfig_file_prefix = ""  # You can add a prefix before name - ex.: env_development.ini\nconfig_file_suffix = ""  # You can add a suffix after name - ex.: development_v1.ini\ndefault_environment = "development" # use this if you do not want to define the ENVIRONMENT key\nconfig_file_path = "."  # relative path from project root for configuration files\n```\n\n### How Stela find the Environment Variables?\n\nWhen asked for a value, Stela will try to find a environment variable\nusing the full uppercase slug path. For example, if key is `foo.bar`,\nStela will search for a env called `FOO_BAR` before returning his dict\nvalue. Also, you can add a prefix/suffix in this calculated environment\nname (ex.: `MYPROJECT_FOO_BAR`). To do this, define them in\n`pyproject.toml`:\n\n```toml\n[tools.stela]\nenvironment_prefix = "MYPROJECT_"\nenvironment_suffix = ""\n```\n\nIn above case, Stela will look for the `MYPROJECT_FOO_BAR` env:\n\n```python\n# FOO_BAR = "hello_world" or\n# MYPROJECT_FOO_BAR = "hello world" if you define environment_prefix\nfrom stela import settings\n\nmy_conf = settings["foo.bar"]\n# my_conf = "hello world"\n```\n\nAlso, you can define Stela to never get values from environment\nvariables, only from dictionary:\n\n```toml\n[tools.stela]\ndo_not_read_environment = true\n```\n\n### How Stela handle more complex cases?\n\nStela uses this lifecycle to handle the settings load:\n\n```text\nPre-Load (optional) -> Load or Default Load > Post-Load (optional)\n```\n\nIf you have more complex cases to retrieve your project settings (ex.:\nreading external services, load settings from database or parse from a\ndifferent file format or library), you can use Stela decorators for\n`pre_load`, `load` and `post_load` phases:\n\n* If you use the `pre_load` decorator, data parsed in this phase will be\n  passed to **Load** phase.\n* If you use the `load` decorator, it will be used instead the Stela\'s\n  `default_loader`. The data returned here will update the dictionary\n  received in previous phase\n* If you use the `post_load` decorator, data returned in this phase will\n  update the dictionary received in previous phase.\n\nOnly one function are allowed per phase.\n\n>  These loaders need to return a valid Python dictionary.\n\n##### Lifecycle example:\n\n```python\nfrom stela.decorators import pre_load, load, post_load\nfrom typing import Dict, Any\nfrom stela import StelaOptions\n\n@pre_load\ndef pre_load(options: StelaOptions) -> Dict[Any, Any]:\n    # data value is: {}\n    # Stela Options are available in options object.\n    return {"foo": "bar"}\n\n@load\ndef load(data: dict, options: StelaOptions) -> Dict[Any, Any]:\n    # data value is: {"foo": "bar"}\n    # Using load, default_loader will not be invoked\n    return {"has_dogs": True}\n\n@post_load\ndef post_load(data: dict, options: StelaOptions) -> Dict[Any, Any]:\n    # data value is: {"foo": "bar", "has_dogs": True}\n    return {"number_of_dogs": 1}\n\n# Final data is {"foo": "bar", "has_dogs": True, "number_of_dogs": 1}\n```\n\n##### Post-Load SSM Example:\n\n```python\n# stela_loader.py\nimport boto3\nfrom stela.decorators import post_load\nfrom stela import StelaOptions\nfrom typing import Dict, Any\n\n@post_load\ndef add_ssm_parameters(data: dict, options: StelaOptions) -> Dict[Any, Any]:\n    """Load settings from SSM to current Stela data.\n    \n    This loader will be called after Stela default loader.\n    Data returned must be a Python Dictionary.\n    \n    :param data (dict): Data parsed from previous phases\n    :param options (obj): Stela Options from pyproject.toml\n    :return Dict[Any, Any]\n    """\n    ssm = boto3.client(\'ssm\')\n    environment = options.current_environment\n    parameters = ssm.get_parameters_by_path(\n        Name=f\'/foo/bar/{environment}\',\n        WithDecryption=True\n    )\n    return parameters\n```\n\n### When Stela read the data?\n\nStela are imported once, at module level. This is the python equivalent\nfor a Singleton object - project settings are load and immediately\navailable:\n\n```python\nfrom stela import settings\nfrom flask import Flask\n\napp = Flask(__name__)\napp.config.update(\n    SECRET_KEY=settings["my_app.secret"] # will read from dict or MY_APP_SECRET value\n)\n\n@app.route("/")\ndef hello():\n    return f"Hello, Environment is {settings.stela_options.current_environment}"\n```\n\nIf you need to reload settings, use the `stela.stela_reload` function:\n\n```python\nfrom stela import stela_reload\n\ndef test_different_environments(monkeypatch):\n    from stela import settings\n    assert settings.stela_options.current_environment == "test"\n    \n    monkeypatch.setenv("ENVIRONMENT", "production")\n    settings = stela_reload()\n    assert settings.stela_options.current_environment == "production"\n    monkeypatch.delenv("ENVIRONMENT")\n```\n\n### How Stela read the dictionary values?\n\nStela will respect the file format limitations. For example, INI files\nalways return values as string, TOML files returning datetime objects,\netc...\n\nFor environment variables, Stela will return value as string, by\ndefault. For example: `NUMBER_OF_CATS=3` will return a string.\n\nYou can set Stela to literal evaluate these values, as per\n[ast.literal_eval](https://docs.python.org/3.7/library/ast.html?highlight=literal_eval#ast.literal_eval)\nrules. To do this, add in `pyproject.toml`:\n\n```toml\n[tool.stela]\nevaluate_data = true\n```\n\n### All Stela Configuration Options:\n\nAll configuration files can be override using a environment variable,\nusing the `STELA_` prefix. For example, `default_environment` turns\n`STELA_DEFAULT_ENVIRONMENT`.\n\n```toml\n# All keys and default values available in pyproject.toml\n[tool.stela]\nenvironment_variable_name = "ENVIRONMENT"\nconfig_file_extension = "INI" # YAML, TOML, JSON\nconfig_file_prefix = ""  # You can add a prefix before name - ex.: env_development.ini\nconfig_file_suffix = ""  # You can add a suffix after name - ex.: development_v1.ini\nconfig_file_path = "."\nenvironment_prefix = ""  # ex.: settings["foo.bar"\'] looks for MY_PREFIX_FOO_BAR\nenvironment_suffix = ""  # ex.: settings["foo.bar"\'] looks for FOO_BAR_MY_SUFFIX\ndefault_environment = ""\nevaluate_data = false\ndo_not_read_environment = false\nshow_logs = true  # as per loguru settings.\n```\n\n### Not working?\n\nDont panic. Get a towel and, please, open a\n[issue](https://github.com/chrismaille/stela/issues).\n',
    'author': 'Chris Maillefaud',
    'author_email': 'chrismaillefaud@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/chrismaille/stela',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
